{- Copyright (c) 2015, Peter Ljung <ljung.peter@gmail.com>

Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
granted, provided that the above copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
USE OR PERFORMANCE OF THIS SOFTWARE. -}

module StockIndicators where

import Prelude as P
import TimeSeries as TS

len :: (Floating b) => [a] -> b
len xs = fromIntegral $ P.length xs

mean :: (Floating a) => [a] -> a
mean xs = sum xs / len xs

variance :: (Floating a) => [a] -> a
variance xs = (1/len xs) * (sum $ P.map (\x -> (x-mx)*(x-mx)) xs) where
  mx = mean xs

covariance :: (Floating a) => ([a], [a]) -> a
covariance (xs, ys) = (1/len xs) * (sum $ P.map (\(x, y) -> (x-mx)*(y-my)) $ zip xs ys) where
  mx = mean xs
  my = mean ys

stddev :: (Floating a) => [a] -> a
stddev xs = sqrt $ variance xs

-- Calculate moving average for a time series over n number of data points
movingAvarage :: (Floating a) => Int -> TimeSeries a -> TimeSeries a
movingAvarage n ts = TS.map mean $ slidingWindow n ts

-- Calculate variance for a time series over n number of data points
movingVariance :: (Floating a) => Int -> TimeSeries a -> TimeSeries a
movingVariance n ts = TS.map variance $ slidingWindow n ts

-- Calculate standard deviation for a time series over n number of data points
movingDeviation :: (Floating a) => Int -> TimeSeries a -> TimeSeries a
movingDeviation n ts = TS.map stddev $ slidingWindow n ts

-- Calculate co-variance for a time series over n number of data points
-- Ref: <https://en.wikipedia.org/wiki/Covariance#Calculating_the_sample_covariance>
movingCoVariance :: (Floating a) => Int -> (TimeSeries a, TimeSeries a) -> TimeSeries a
movingCoVariance n (xs, ys) = TS.map covariance $ combine (,) (slidingWindow n xs) (slidingWindow n ys)

-- Calculate beta-regression for a time series over n number of data points
-- Ref: <https://en.wikipedia.org/wiki/Beta_(finance)>
-- Note: filter is required because variance may be zero
movingBeta :: (RealFloat a) => Int -> (TimeSeries a, TimeSeries a) -> TimeSeries a
movingBeta n (xs, ys) = TS.filter (\e -> not $ isNaN e)
                      $ combine (/) (movingCoVariance n (xs, ys)) (movingVariance n ys)

-- Determine when time series a cross from below of b from previous value to current
crossesBelow :: (Num a, Ord a) => TimeSeries a -> TimeSeries a -> Signal
crossesBelow a b = mkSignal id s where
  a' = combine (,) (TS.delay 1 a) a
  b' = combine (,) (TS.delay 1 b) b
  s = combine crossBelow a' b'
  -- Does (a_prev,a) pass (b_prev,b) from below i.e. is a_prev < b_prev && a > b ?
  crossBelow :: (Num a, Ord a) => (a,a) -> (a,a) -> Bool
  crossBelow (a',a) (b',b) | a' < b' && a > b = True
                           | otherwise        = False

crossesAbove :: (Num a, Ord a) => TimeSeries a -> TimeSeries a -> Signal
crossesAbove = flip crossesBelow

-- Derivative of time series
-- TODO: Make use of dt of time series
differentiate :: (Floating a) => TimeSeries a -> TimeSeries a
differentiate a = combine (\a b -> ((a-b)/h)) (TS.delay 1 a) a where h = 1

-- Trails max of time series
trailingMax :: (Num a, Ord a) => TimeSeries a -> TimeSeries a
trailingMax ts@(TS _ _ as) = smap tmax (head as) ts where
  tmax s e = let r = max s e in (r, r)

-- Trails min of time series
trailingMin :: (Num a, Ord a) => TimeSeries a -> TimeSeries a
trailingMin ts@(TS _ _ as) = smap tmin (head as) ts where
  tmin s e = let r = min s e in (r, r)

-- Stop trail indicator where p is the percentage above/below min/max trail
-- according to <http://stocata.org/metastock/stop_trail_perc.html>
stopTrail :: (Num a, Ord a) => a -> TimeSeries a -> TimeSeries a
stopTrail p ts@(TS _ _ as) = smap stopFunc s0 ts where
  s0 = (True, head as)    -- start with barrier
  stopFunc (True,  v)  e | e > v*(1-p) = let r = max e v in ((True,  r), r*(1-p))
                         | otherwise   = let r = e       in ((False, r), r*(1+p))
  stopFunc (False, v)  e | e < v*(1+p) = let r = min e v in ((False, r), r*(1+p))
                         | otherwise   = let r = e       in ((True,  r), r*(1-p))
