{- Copyright (c) 2015, Peter Ljung <ljung.peter@gmail.com>

Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby
granted, provided that the above copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
USE OR PERFORMANCE OF THIS SOFTWARE. -}

module TimeSeries where

import Prelude as P
import Fifo as F
-- import Data.List (foldl')
import Data.Time.Calendar as D
import Data.Time.Clock as C
import Data.Time.LocalTime

type Index = Int    -- Time series data index 0..(n-1)
-- Time series. Time series data is represented as a list of chronological data items
-- First item is at `start` time. Time distance between items are `dt`
data TimeSeries a = TS { start :: D.Day
                       , dt :: C.DiffTime
                       , series :: [a] } deriving (Show)
type Signal       = TimeSeries Bool

aDay :: C.DiffTime
aDay = C.secondsToDiffTime (60*60*24)

mkEmpty :: TimeSeries a
mkEmpty = TS nullDay aDay [] where
  nullDay = D.fromGregorian 0 0 0

isEmpty :: TimeSeries a -> Bool
isEmpty (TS _ _ as) = null as

length :: TimeSeries a -> Int
length (TS _ _ as) = P.length as

-- Map for time series
map :: (a -> b) -> (TimeSeries a) -> (TimeSeries b)
map f ts@(TS _ _ as) = ts { series = P.map f as } where

-- Take for time series
take :: Int -> (TimeSeries a) -> (TimeSeries a)
take n ts@(TS _ _ as) = ts { series = P.take n as } where

-- Drop for time series
drop :: Int -> (TimeSeries a) -> (TimeSeries a)
drop n ts@(TS _ _ as) = ts { series = P.drop n as } where

-- Stateful map functions for time series. It runs through series from chronologically, while
-- keeping a state through the process. Map function :: state -> elem -> (state', elem')
smap :: (s -> a -> (s, b)) -> s -> TimeSeries a -> TimeSeries b
smap f s0 ts@(TS _ _ as) = ts { series = as' } where
  as' = reverse $ snd $ foldl smap' (s0, []) as
  smap' (s, es') e = let (s', e') = f s e in (s', e':es')

-- Delay time series by a number of steps. Fill beginning with zeroes.
delay :: (Num a) => Int -> TimeSeries a -> TimeSeries a
delay n ts@(TS _ _ as) = ts { series = as' } where
  as' = P.take (P.length as) $ replicate n 0 ++ as

-- Filter time series i.e. pass values that satify predicate or keep previous passed value
-- Previous value is kept as state. Initial previous value is zero
filter :: (Num a) => (a -> Bool) -> TimeSeries a -> TimeSeries a
filter f ts = smap filt s0 ts where
  s0 = 0
  filt s e = let s' = if f e then e else s in (s', s')

-- Combine two time series. Note that time range/scale must match
combine :: (a -> a -> b) -> TimeSeries a -> TimeSeries a -> TimeSeries b
combine f (TS s dt as) (TS s' dt' bs)
    | valid     = TS s dt $ P.zipWith f as bs
    | otherwise = error "time series time scale must match" where
  valid = (s, dt, P.length as) == (s', dt', P.length bs)

-- Make signal from time series
mkSignal :: (a -> Bool) -> TimeSeries a -> Signal
mkSignal = TimeSeries.map

-- Make sliding window from time series. Use a fifo as state
slidingWindow :: Int -> TimeSeries a -> TimeSeries [a]
slidingWindow n ts = smap window s0 ts where
  s0 = F.mkEmpty n
  window s e = let s' = push s e in (s', F.toList s')

-- Convert a time series to time list of time series values
-- TODO: dt is ignored. 1 day is used
asTimed :: TimeSeries a -> [(LocalTime, a)]
asTimed (TS s _ as) = P.zipWith ixToTime [0..] as where
  ixToTime :: Int -> a -> (LocalTime, a)
  ixToTime i v = (LocalTime (addDays (fromIntegral i) s) midnight, v)
